"""Descriptor to be taken from datapoint (e.g. average) that can be tested with a `~qualpipe.core.criterion`."""

from abc import ABC, abstractmethod

import numpy as np
from ctapipe.core import Component, TelescopeComponent

__all__ = [
    "BaseDescriptor",
    "BaseTelescopeDescriptor",
    "BasePercentageDescriptor",
    "BaseMedianDescriptor",
    "BaseMeanDescriptor",
    "PercentageDescriptor",
    "MedianDescriptor",
    "MeanDescriptor",
    "TelescopePercentageDescriptor",
    "TelescopeMedianDescriptor",
    "TelescopeMeanDescriptor",
]


class BaseDescriptor(ABC):
    """Base class for all descriptors."""

    @abstractmethod
    def __call__(self, datapoint, tel_id=None):
        """Return the modified datapoint value."""
        pass

    def __str__(self):
        """Pretty string representation for printing."""
        trait_names = self.trait_names()
        formatted_attrs = "\n  ".join(
            f"{name}: {getattr(self, name)}" for name in trait_names
        )
        return f"{self.__class__.__name__}:\n  {formatted_attrs}"


class BaseTelescopeDescriptor(TelescopeComponent):
    """Base class for telescope-specific descriptors."""

    @abstractmethod
    def __call__(self, datapoint, tel_id):
        """Return the modified datapoint value for a specific telescope."""
        pass


class BasePercentageDescriptor(BaseDescriptor):
    """Base class for percentage descriptors."""

    def __call__(self, datapoint, tel_id=None):
        """Compute descriptor value."""
        value = np.array(datapoint)
        return (value != 0).sum() / len(value) * 100


class BaseMedianDescriptor(BaseDescriptor):
    """Base class for median descriptors."""

    def __call__(self, datapoint, tel_id=None):
        """Compute descriptor value."""
        return float(np.nanmedian(datapoint))


class BaseMeanDescriptor(BaseDescriptor):
    """Base class for mean descriptors."""

    def __call__(self, datapoint, tel_id=None):
        """Compute descriptor value."""
        return float(np.nanmean(datapoint))


class PercentageDescriptor(BasePercentageDescriptor, Component):
    """Percentage of non-zero values in an array-like metric."""

    pass


class MedianDescriptor(BaseMedianDescriptor, Component):
    """NaN-Median of an array."""

    pass


class MeanDescriptor(BaseMeanDescriptor, Component):
    """NaN-Mean of an array."""

    pass


class TelescopePercentageDescriptor(BasePercentageDescriptor, BaseTelescopeDescriptor):
    """Percentage of non-zero values in an array-like metric, specific to a telescope."""

    pass


class TelescopeMedianDescriptor(BaseMedianDescriptor, BaseTelescopeDescriptor):
    """NaN-Median of an array, specific to a telescope."""

    pass


class TelescopeMeanDescriptor(BaseMeanDescriptor, BaseTelescopeDescriptor):
    """NaN-Mean of an array, specific to a telescope."""

    pass



"""
Configuration file for the Sphinx documentation builder.

This file only contains a selection of the most common options. For a full
list see the documentation:
https://www.sphinx-doc.org/en/master/usage/configuration.html
"""

# -- Project information -----------------------------------------------------
# import your own package here
import qualpipe

project = "qualpipe"
copyright = "2023, UniGE-UniBern-CTAO DPPS QualPipe Group"
author = "UniGE-UniBern-CTAO DPPS QualPipe Group"
version = qualpipe.__version__
# The full version, including alpha/beta/rc tags.
release = version


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.githubpages",
    "sphinx.ext.intersphinx",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "numpydoc",
    "sphinx_changelog",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.viewcode",
    "sphinx.ext.mathjax",
    "sphinx_paramlinks",
    "myst_parser",
    "sphinxarg.ext",
    "nbsphinx",
]

myst_enable_extensions = [
    "linkify",
]

myst_heading_anchors = 3

# Default options for autodoc directives
autodoc_default_options = {
    "members": True,
    "member-order": "bysource",
    "special-members": "__init__",
    "undoc-members": True,
    "exclude-members": "__weakref__",
}

# intersphinx allows referencing other packages sphinx docs
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "astropy": ("https://docs.astropy.org/en/stable/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "ctapipe": ("https://ctapipe.readthedocs.io/en/stable/", None),
}

nitpick_ignore = [
    ("py:class", "StrDict"),
    ("py:class", "ClassesType"),
    ("py:obj", "parent=self"),
]

# autosectionlabel_prefix_document = True

# Generate todo blocks
todo_include_todos = True

# Add any paths that contain templates here, relative to this directory.
templates_path = []

# The suffix of source filenames.
source_suffix = {
    ".rst": "restructuredtext",
    ".md": "markdown",
}

# The master toctree document.
master_doc = "index"

# Ignore example notebook errors
nbsphinx_allow_errors = True
nbsphinx__timeout = 200  # allow max 2 minutes to build each notebook

numpydoc_show_class_members = False

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["changes"]

# have all links automatically associated with the right domain.
default_role = "py:obj"


# -- Options for HTML output -------------------------------------------------

html_theme = "ctao"
html_theme_options = dict(
    navigation_with_keys=False,
    logo=dict(text="QualPipe"),
    # setup for displaying multiple versions, also see setup in .gitlab-ci.yml
    switcher=dict(
        json_url="http://cta-computing.gitlab-pages.cta-observatory.org/dpps/qualpipe/qualpipe/versions.json",  # noqa: E501
        version_match="latest" if ".dev" in version else f"v{version}",
    ),
    navbar_center=["version-switcher", "navbar-nav"],
)

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]



"""Defines the Metric class for evaluating data quality metrics."""

from enum import Enum

from ctapipe.core import Component, TelescopeComponent
from ctapipe.core.traits import Bool, List, Unicode
from ctapipe.core.traits import Enum as TraitEnum
from traitlets.config import Config

from .criterion import *  # noqa: F403
from .descriptor import *  # noqa: F403

__all__ = ["ArrayElement", "DataCategory", "Metric"]


class ArrayElement(Enum):
    """Enumeration of array elements."""

    DUMMY = "DUMMY"
    SUBARRAY = "SUBARRAY"
    LST = "LST"
    MST = "MST"
    SST = "SST"
    FRAM = "FRAM"
    LIDAR = "LIDAR"
    WEATHER_STATION = "WEATHER_STATION"


class DataCategory(Enum):
    """Enumeration of data categories."""

    DL0_EVENT = "DL0_EVENT"
    DL1_EVENT = "DL1_EVENT"
    DL2_EVENT = "DL2_EVENT"
    DL3_EVENT = "DL3_EVENT"
    DL0_SERVICE = "DL0_SERVICE"
    DL1_SERVICE = "DL1_SERVICE"
    DL2_SERVICE = "DL2_SERVICE"
    DL3_SERVICE = "DL3_SERVICE"
    DL0_MONITORING = "DL0_MONITORING"
    DL1_MONITORING = "DL1_MONITORING"
    DL2_MONITORING = "DL2_MONITORING"
    DL3_MONITORING = "DL3_MONITORING"


class Metric(Component):
    """Defines a data quality metric that applies a descriptor to input data and evaluates quality criteria."""

    name = Unicode(help="Metric name").tag(config=True)
    input_source = Unicode(
        help="HDF5 table path and column name, e.g., '/table/column'"
    ).tag(config=True)
    array_element = TraitEnum(
        ArrayElement,
        help="Array element",
    ).tag(config=True)
    data_category = TraitEnum(
        DataCategory,
        help="Data category (DataLevel/DataType)",
    ).tag(config=True)
    telescope_specific = Bool(
        False, help="Flag to indicate if this is a telescope-specific metric"
    ).tag(config=True)
    descriptor = Unicode(
        None,
        allow_none=True,
        help="Configuration for the descriptor",
    ).tag(config=True)
    criteria = List(help="List of configurations for quality criteria").tag(config=True)

    def __init__(self, **kwargs):
        self.subarray = kwargs.pop("subarray", None)
        super().__init__(**kwargs)

        # Load descriptor (trivial if none is provided)
        if self.telescope_specific:
            base_component_class = TelescopeComponent
            kwargs_dict = {"subarray": self.subarray}
        else:
            base_component_class = Component
            kwargs_dict = {}

        if self.descriptor:
            self._descriptor = base_component_class.from_name(
                self.descriptor, parent=self, **kwargs_dict
            )
        else:
            self._descriptor = (
                lambda data, tel_id=None: data
            )  # Trivial descriptor (identity function)

        # Load quality criteria
        self._criteria = [
            base_component_class.from_name(
                next(iter(criterion)), config=Config(criterion), **kwargs_dict
            )
            for criterion in self.criteria
        ]
        self.log.debug("Metric %s initialized", self.name)
        self.log.debug("Metric's descriptor: %s", self.descriptor)
        self.log.debug("Metric's criteria:")
        for criterion in self._criteria:
            self.log.debug(criterion)

    def __call__(self, data, tel_id=None):
        """Compute the metric value using the descriptor and evaluate quality criteria."""
        if self.telescope_specific and tel_id is None:
            raise ValueError("Telescope-specific metric requires a telescope ID")
        transformed_data = self._descriptor(data, tel_id)

        results = {
            criterion.__class__.__name__: {
                "config": criterion.get_current_config(),
                "result": criterion(
                    transformed_data, tel_id if self.telescope_specific else None
                ),
            }
            for criterion in self._criteria
        }
        return results



"""Criteria to assess the data quality."""

from abc import ABC, abstractmethod

import numpy as np
from ctapipe.core import Component, TelescopeComponent
from ctapipe.core.traits import Bool, Float, TelescopeParameter

__all__ = [
    "BaseQualityCriterion",
    "BaseTelescopeQualityCriterion",
    "BaseRangeCriterion",
    "BaseThresholdCriterion",
    "RangeCriterion",
    "TelescopeRangeCriterion",
    "ThresholdCriterion",
    "TelescopeThresholdCriterion",
]


class BaseQualityCriterion(ABC):
    """Base class for all quality criteria."""

    @abstractmethod
    def _get_parameters(self, tel_id=None):
        """Retrieve the parameters for the criterion, either globally or per telescope."""
        pass

    @abstractmethod
    def __call__(self, metric, tel_id=None):
        """Evaluate the criterion against the metric."""
        pass

    def __str__(self):
        """Pretty string representation for printing."""
        trait_names = self.trait_names()
        formatted_attrs = "\n  ".join(
            f"{name}: {getattr(self, name)}" for name in trait_names
        )
        return f"{self.__class__.__name__}:\n  {formatted_attrs}"


class BaseTelescopeQualityCriterion(TelescopeComponent):
    """Base class for telescope-specific quality criteria."""

    @abstractmethod
    def __call__(self, metric, tel_id):
        """Evaluate the criterion against the metric for a specific telescope."""
        pass


class BaseRangeCriterion(BaseQualityCriterion):
    """Abstract base class for range-based quality criteria."""

    def _get_parameters(self, tel_id=None):
        """Retrieve the min and max values, either globally or per telescope."""
        if tel_id is not None:
            return self.min_value.tel[tel_id], self.max_value.tel[tel_id]
        return self.min_value, self.max_value

    def __call__(self, data, tel_id=None):
        """Check if all values in the metric are within the range."""
        min_value, max_value = self._get_parameters(tel_id)
        return bool(np.all((data >= min_value) & (data <= max_value)))


class RangeCriterion(BaseRangeCriterion, Component):
    """Range-based quality criterion for general components."""

    min_value = Float(
        help="Minimum allowed value", allow_none=False, default_value=None
    ).tag(config=True)
    max_value = Float(
        help="Maximum allowed value", allow_none=False, default_value=None
    ).tag(config=True)


class TelescopeRangeCriterion(BaseRangeCriterion, BaseTelescopeQualityCriterion):
    """Telescope-specific range criterion using per-telescope parameters."""

    min_value = TelescopeParameter(trait=Float(), help="Minimum allowed value").tag(
        config=True
    )
    max_value = TelescopeParameter(trait=Float(), help="Maximum allowed value").tag(
        config=True
    )


class BaseThresholdCriterion(BaseQualityCriterion):
    """Abstract base class for threshold-based criteria."""

    def _get_parameters(self, tel_id=None):
        """Retrieve the threshold value, either globally or per telescope."""
        return self.threshold.tel[tel_id] if tel_id is not None else self.threshold

    def __call__(self, data, tel_id=None):
        """Check if all values in the metric satisfy the threshold condition."""
        threshold = self._get_parameters(tel_id)
        return bool(
            np.all(data > threshold) if self.above else np.all(data < threshold)
        )


class ThresholdCriterion(BaseThresholdCriterion, Component):
    """Threshold-based quality criterion for general components."""

    threshold = Float(help="Threshold value", allow_none=False, default_value=None).tag(
        config=True
    )
    above = Bool(
        help="Check if values are above (True) or below (False) the threshold",
        allow_none=False,
        default_value=None,
    ).tag(config=True)


class TelescopeThresholdCriterion(
    BaseThresholdCriterion, BaseTelescopeQualityCriterion
):
    """Telescope-specific threshold criterion using per-telescope parameters."""

    threshold = TelescopeParameter(trait=Float(), help="Threshold value").tag(
        config=True
    )
    above = Bool(
        help="Check if values are above (True) or below (False) the threshold",
        allow_none=False,
        default_value=None,
    ).tag(config=True)



import pytest
from qualpipe.core.descriptor import (
    MeanDescriptor,
    MedianDescriptor,
    PercentageDescriptor,
    TelescopeMeanDescriptor,
    TelescopeMedianDescriptor,
    TelescopePercentageDescriptor,
)


@pytest.mark.verifies_usecase("UC-140-1.2.1")
def test_median_descriptor():
    datapoint = [0, 1, 2, 3, 4]
    descriptor = MedianDescriptor()
    assert descriptor(datapoint) == 2


@pytest.mark.verifies_usecase("UC-140-1.2.1")
def test_percentage_descriptor():
    datapoint = [0, 0, 0, 1, 0]
    descriptor = PercentageDescriptor()
    assert descriptor(datapoint) == 20  # %


@pytest.mark.verifies_usecase("UC-140-1.2.1")
def test_mean_descriptor():
    datapoint = [0, 1, 2, 3, 4]
    descriptor = MeanDescriptor()
    assert descriptor(datapoint) == 2


@pytest.mark.verifies_usecase("UC-140-1.2.1")
def test_telescope_median_descriptor():
    datapoint = [0, 1, 2, 3, 4]
    descriptor = TelescopeMedianDescriptor(subarray=None)
    assert descriptor(datapoint, tel_id=1) == 2


@pytest.mark.verifies_usecase("UC-140-1.2.1")
def test_telescope_percentage_descriptor():
    datapoint = [0, 0, 0, 1, 0]
    descriptor = TelescopePercentageDescriptor(subarray=None)
    assert descriptor(datapoint, tel_id=1) == 20  # %


@pytest.mark.verifies_usecase("UC-140-1.2.1")
def test_telescope_mean_descriptor():
    datapoint = [0, 1, 2, 3, 4]
    descriptor = TelescopeMeanDescriptor(subarray=None)
    assert descriptor(datapoint, tel_id=1) == 2



from pathlib import Path

import pytest
import yaml
from ctapipe.instrument import SubarrayDescription
from qualpipe.core.metric import Metric
from traitlets.config import Config

project_root = Path(__file__).resolve().parents[2]
data_path = project_root / "tests" / "data"
datapoint = [0, 1, 2, 3, 4]


@pytest.fixture()
def setup_metric():
    metric_config_yaml = """
    Metric:
      name: "test_metric"
      input_source: "dummy_data_file.h5:/data"
      array_element: DUMMY
      data_category: DL1_EVENT
      telescope_specific: False
      descriptor: MeanDescriptor
      criteria:
        - RangeCriterion:
            min_value: 0
            max_value: 5
    """
    metric_config_dict = yaml.safe_load(metric_config_yaml)
    metric_config = Config(metric_config_dict)
    metric = Metric(config=metric_config)
    return metric


@pytest.mark.verifies_usecase("UC-140-1.2.1")
def test_metric_descriptor(setup_metric):
    metric = setup_metric
    result = metric._descriptor(datapoint)
    assert result == 2


@pytest.mark.verifies_usecase("UC-140-1.2.1")
def test_metric_apply_criteria(setup_metric):
    metric = setup_metric
    results = metric(datapoint)
    assert "RangeCriterion" in results
    assert results["RangeCriterion"]["result"]


@pytest.fixture()
def setup_telescope_specific_metric():
    subarray = SubarrayDescription.from_hdf(f"{data_path}/MonteCarloArray.hdf")
    metric_config_yaml = """
    Metric:
      name: "test_telescope_metric"
      input_source: "dummy_data_file.h5:/data"
      array_element: LST
      data_category: DL1_EVENT
      telescope_specific: True
      descriptor: TelescopeMeanDescriptor
      criteria:
        - TelescopeRangeCriterion:
            min_value:
              - [type, LST*, 0]
              - [type, MST*, 5]
            max_value:
              - [type, LST*, 10]
              - [type, MST*, 15]
    """
    metric_config_dict = yaml.safe_load(metric_config_yaml)
    metric_config = Config(metric_config_dict)
    metric = Metric(config=metric_config, subarray=subarray)
    return metric


@pytest.mark.verifies_usecase("UC-140-1.2.1")
def test_telescope_specific_metric_descriptor(setup_telescope_specific_metric):
    metric = setup_telescope_specific_metric
    result = metric._descriptor(datapoint, tel_id=1)
    assert result == 2


@pytest.mark.verifies_usecase("UC-140-1.2.1")
def test_telescope_specific_metric_apply_criteria(setup_telescope_specific_metric):
    metric = setup_telescope_specific_metric
    results = metric(datapoint, tel_id=1)
    assert "TelescopeRangeCriterion" in results
    assert results["TelescopeRangeCriterion"]["result"]


@pytest.fixture()
def setup_metric_with_multiple_criteria():
    metric_config_yaml = """
    Metric:
      name: "test_metric_multiple_criteria"
      input_source: "dummy_data_file.h5:/data"
      array_element: DUMMY
      data_category: DL1_EVENT
      telescope_specific: False
      descriptor: MeanDescriptor
      criteria:
        - RangeCriterion:
            min_value: 0
            max_value: 5
        - ThresholdCriterion:
            threshold: 2
            above: True
    """
    metric_config_dict = yaml.safe_load(metric_config_yaml)
    metric_config = Config(metric_config_dict)
    metric = Metric(config=metric_config)
    return metric


@pytest.mark.verifies_usecase("UC-140-1.2.1")
def test_metric_with_multiple_criteria_descriptor(setup_metric_with_multiple_criteria):
    metric = setup_metric_with_multiple_criteria
    result = metric._descriptor(datapoint)
    assert result == 2


@pytest.mark.verifies_usecase("UC-140-1.2.1")
def test_metric_with_multiple_criteria_apply_criteria(
    setup_metric_with_multiple_criteria,
):
    metric = setup_metric_with_multiple_criteria
    results = metric(datapoint)
    assert "RangeCriterion" in results
    assert results["RangeCriterion"]["result"]
    assert "ThresholdCriterion" in results
    assert not results["ThresholdCriterion"]["result"]



"""Main tool for QualPipe backend that coordinates the reading of the metrics and application of criteria."""

import json

from ctapipe.core import Tool
from ctapipe.core.telescope_component import TelescopePatternList
from ctapipe.core.traits import List, Path, Unicode
from ctapipe.instrument import SubarrayDescription
from ctapipe.io import read_table
from traitlets.config import Config

from ..core.metric import Metric


class CustomJSONEncoder(json.JSONEncoder):
    """Encoder for TelescopePatternList."""

    def default(self, obj):
        """Convert TelescopePatternList to a list of tuples."""
        if isinstance(obj, TelescopePatternList):
            # Convert TelescopePatternList to a list of tuples
            return list(obj)
        return super().default(obj)


class MetricEvaluator(Tool):
    """
    Main class to run the automatic QualPipe.

    The MetricEvaluator class is responsible for managing the sequence of operations required to
    prepare metrics and criteria according to the configuration files. It coordinates the retrieval,
    computation, and application of metrics and criteria within QualPipe.


    :Example:

    Configuration example:

    .. code-block:: yaml

        MetricEvaluator:
          input_url: "dummy_data_file.h5"
          metrics:
            - name: my_array_measurement                # this name will be used in report to identify the measurement
              input_source: "/my_table/my_data_array"   # path to the data in the HDF5 file
              array_element: LST
              data_category: DL1_EVENT
              telescope_specific: False
              descriptor: MeanDescriptor
              criteria:
                - RangeCriterion:
                    min_value: 0
                    max_value: 5

            - name: my_float_measurement
              input_source: "/my_table/my_data_float"
              array_element: LST
              data_category: DL1_EVENT
              telescope_specific: False
              descriptor: Descriptor
              criteria:
                - RangeCriterion:
                    min_value: 0.3
                    max_value: 0.4
          output_url: "report.json"
    """

    input_url = Path(
        help="URL of the input file containing monitoring data",
        allow_none=False,
        exists=True,
        directory_ok=False,
        file_ok=True,
    ).tag(config=True)
    metrics = List(help="List of configs for metrics", minlen=1).tag(config=True)
    output_url = Unicode(
        help="URL of the output report file",
        allow_none=False,
    ).tag(config=True)

    aliases = {
        "input-url": "MetricEvaluator.input_url",
        "output-url": "MetricEvaluator.output_url",
    }

    MONITORING_TEL_GROUP = "/dl1/monitoring/telescope/"

    classes = [Metric]

    def setup(self):
        """Set up the MetricEvaluator class."""
        self._metrics = []
        for metric_dict in self.metrics:
            self.log.debug("MetricEvaluator : Loading metric %s", metric_dict)
            metric_cfg = Config(metric_dict)
            if "telescope_specific" in metric_dict["Metric"]:
                subarray = SubarrayDescription.from_hdf(self.input_url)
                self._metrics.append(
                    Metric(
                        config=Config(metric_cfg),
                        subarray=subarray,
                    )
                )
            else:
                self._metrics.append(Metric(Config(metric_cfg)))

    def _read_data(self, metric, tel_id=None):
        """Read data from the input source."""
        # Implement the logic to read data from the input source
        # For the moment, only DL1 data following the reference DL1 implementation in ctapipe is supported.
        if "DL1" not in metric.data_category:
            raise NotImplementedError("Only DL1 data is supported for the moment")

        table_path, column = metric.input_source.rsplit("/", 1)
        if metric.telescope_specific:
            if tel_id is None:
                raise ValueError("Telescope-specific metric requires a telescope ID")
            data = read_table(
                self.input_url,
                f"{table_path}/tel_{tel_id:03d}",
            )
        else:
            data = read_table(self.input_url, f"{table_path}")

        return data[column]

    def start(self):
        """Loop over all metrics and apply criteria."""
        self.results = {}
        for metric in self._metrics:
            self.results[metric.name] = {}
            if metric.telescope_specific:
                for tel_id in metric.subarray.tel_ids:
                    data = self._read_data(metric, tel_id)
                    results = metric(data, tel_id)
                    id_info = f"{metric.array_element}-{tel_id:03d}"
                    self.results[metric.name][id_info] = results
            else:
                data = self._read_data(metric)
                results = metric(data)
                self.results[metric.name][metric.array_element] = results

    def finish(self):
        """Finish the job and write the final report."""
        self.log.debug("MetricEvaluator : Results: %s", self.results)
        with open(self.output_url, "w") as f:
            json.dump(self.results, f, cls=CustomJSONEncoder, indent=4)
        self.log.info("MetricEvaluator : Finish running automatic QualPipe")


def main():
    """Run the MetricEvaluator tool."""
    tool = MetricEvaluator()
    tool.run()



import json
import os

import pytest
from ctapipe.core import run_tool
from qualpipe.tools import MetricEvaluator


@pytest.fixture()
def config_file(tmp_path):
    config_content = """
    MetricEvaluator:
      metrics:
        - Metric:
            name: "atmosphere_height_measurement"
            input_source: "/simulation/service/atmosphere_density_profile/height"
            array_element: WEATHER_STATION
            data_category: DL1_EVENT
            telescope_specific: False
            criteria:
              - RangeCriterion:
                  min_value: 0.0
                  max_value: 120.0

        - Metric:
            name: "atmosphere_height_mean"
            input_source: "/simulation/service/atmosphere_density_profile/height"
            array_element: WEATHER_STATION
            data_category: DL1_EVENT
            telescope_specific: False
            descriptor: MeanDescriptor
            criteria:
              - RangeCriterion:
                  min_value: 30
                  max_value: 40

        - Metric:
            name: "obs_id"
            input_source: "/dl1/event/telescope/parameters/obs_id"
            array_element: LST
            data_category: DL1_EVENT
            telescope_specific: True
            criteria:
              - TelescopeRangeCriterion:
                  min_value: 98
                  max_value: 99
    """
    config_path = tmp_path / "config.yaml"
    config_path.write_text(config_content)
    return config_path


@pytest.fixture()
def dummy_data_file(tmp_path):
    return os.path.join(
        os.path.dirname(__file__), "../data/gamma_LSTSubarray_run99.dl1.h5"
    )


@pytest.mark.xfail(
    reason="This test is expected to fail in CI due to the absence of the test file. A proper test file will be added later."
)
def test_metric_evaluator(config_file, dummy_data_file, tmp_path):
    output_report_path = "report.json"

    # Run the MetricEvaluator tool
    run_tool(
        MetricEvaluator(),
        argv=[
            f"--config={config_file}",
            f"--input-url={dummy_data_file}",
            f"--output-url={output_report_path}",
        ],
    )

    # Check if the output report file is created
    assert os.path.exists(output_report_path)

    # Load the report and check its content
    with open(output_report_path) as f:
        report = json.load(f)

    assert "atmosphere_height_measurement" in report
    assert "atmosphere_height_mean" in report
    assert "obs_id" in report
    # Add more assertions based on the expected content of the report



from pathlib import Path  # noqa: I001

import numpy as np
import pytest
import yaml
from ctapipe.core import Component, TelescopeComponent
from ctapipe.instrument import SubarrayDescription

from qualpipe.core.criterion import (
    RangeCriterion,  # noqa: F401
    TelescopeRangeCriterion,  # noqa: F401
    ThresholdCriterion,  # noqa: F401
    TelescopeThresholdCriterion,  # noqa: F401
)  # noqa: F401
from traitlets.config import Config
from traitlets import TraitError

project_root = Path(__file__).resolve().parents[2]
data_path = project_root / "tests" / "data"


@pytest.fixture()
def range_criterion():
    criterion_config_yaml = """
    RangeCriterion:
      min_value: 0
      max_value: 15
    """
    criterion_config_dict = yaml.safe_load(criterion_config_yaml)
    criterion_config = Config(criterion_config_dict)
    criterion = Component.from_name(
        list(criterion_config.keys())[0], config=criterion_config
    )
    return criterion


@pytest.fixture()
def telescope_range_criterion():
    subarray = SubarrayDescription.from_hdf(f"{data_path}/MonteCarloArray.hdf")
    criterion_config_yaml = """
    TelescopeRangeCriterion:
      min_value:
        - [type, LST*, 3.]
        - [type, MST*, 11]
      max_value:
        - [type, LST*, 10]
        - [type, MST*, 20]
    """
    criterion_config_dict = yaml.safe_load(criterion_config_yaml)
    criterion_config = Config(criterion_config_dict)
    criterion = TelescopeComponent.from_name(
        list(criterion_config.keys())[0], subarray, config=criterion_config
    )
    return criterion


@pytest.fixture()
def threshold_criterion():
    criterion_config_yaml = """
    ThresholdCriterion:
      threshold: 10
      above: True
    """
    criterion_config_dict = yaml.safe_load(criterion_config_yaml)
    criterion_config = Config(criterion_config_dict)
    criterion = Component.from_name(
        list(criterion_config.keys())[0], config=criterion_config
    )
    return criterion


@pytest.fixture()
def telescope_threshold_criterion():
    subarray = SubarrayDescription.from_hdf(f"{data_path}/MonteCarloArray.hdf")
    criterion_config_yaml = """
    TelescopeThresholdCriterion:
      threshold:
        - [type, LST*, 10]
        - [type, MST*, 20]
      above: True
    """
    criterion_config_dict = yaml.safe_load(criterion_config_yaml)
    criterion_config = Config(criterion_config_dict)
    criterion = TelescopeComponent.from_name(
        list(criterion_config.keys())[0], subarray, config=criterion_config
    )
    return criterion


@pytest.mark.verifies_usecase("UC-140-1.2.1")
def test_range_criterion_call(range_criterion):
    data_within_range = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    data_outside_range = np.array([16, 17, 18])
    assert range_criterion(data_within_range)
    assert not range_criterion(data_outside_range)


@pytest.mark.verifies_usecase("UC-140-1.2.1")
def test_telescope_range_criterion_call(telescope_range_criterion):
    data_within_range_tel1 = np.array([4, 5, 6, 7, 8, 9])
    data_outside_range_tel1 = np.array([1, 2, 3, 10, 11, 12])
    data_within_range_tel5 = np.array([16, 17, 18, 19])
    data_outside_range_tel5 = np.array([1, 2, 3, 16, 17, 18])
    assert telescope_range_criterion(data_within_range_tel1, tel_id=1)
    assert not telescope_range_criterion(data_outside_range_tel1, tel_id=1)
    assert telescope_range_criterion(data_within_range_tel5, tel_id=5)
    assert not telescope_range_criterion(data_outside_range_tel5, tel_id=5)


@pytest.mark.verifies_usecase("UC-140-1.2.1")
def test_threshold_criterion_call(threshold_criterion):
    data_above_threshold = np.array([11, 12, 13])
    data_below_threshold = np.array([8, 9])
    assert threshold_criterion(data_above_threshold)
    assert not threshold_criterion(data_below_threshold)


@pytest.mark.verifies_usecase("UC-140-1.2.1")
def test_telescope_threshold_criterion_call(telescope_threshold_criterion):
    data_above_threshold_tel1 = np.array([11, 12, 13])
    data_below_threshold_tel1 = np.array([8, 9])
    data_above_threshold_tel5 = np.array([21, 22, 23])
    data_below_threshold_tel5 = np.array([18, 19])
    assert telescope_threshold_criterion(data_above_threshold_tel1, tel_id=1)
    assert not telescope_threshold_criterion(data_below_threshold_tel1, tel_id=1)
    assert telescope_threshold_criterion(data_above_threshold_tel5, tel_id=5)
    assert not telescope_threshold_criterion(data_below_threshold_tel5, tel_id=5)


@pytest.mark.verifies_usecase("UC-140-1.2.1")
def test_missing_min_value_in_range_criterion_config():
    """Missing entirely 'min_value: 0' item."""

    wrong_criterion_config_yaml = """
    RangeCriterion:
      max_value: 15
    """
    wrong_criterion_config_dict = yaml.safe_load(wrong_criterion_config_yaml)
    wrong_criterion_config = Config(wrong_criterion_config_dict)
    criterion = Component.from_name(
        list(wrong_criterion_config.keys())[0], config=wrong_criterion_config
    )
    datapoint = np.array([0, 1, 2, 3, 4])

    # Verify that a TraitError is raisesd when calling the RangeCriterion component
    with pytest.raises(
        TraitError,
    ) as exc_info:
        criterion(datapoint)

    # Access the original cause of the TraitError
    original_exception = exc_info.value.__context__
    assert isinstance(original_exception, KeyError)
    assert "'min_value'" in str(original_exception)


@pytest.mark.verifies_usecase("UC-140-1.2.1")
def test_missing_max_value_in_range_criterion_config():
    """Missing entirely 'max_value: 15' item."""

    wrong_criterion_config_yaml = """
    RangeCriterion:
      min_value: 0
    """
    wrong_criterion_config_dict = yaml.safe_load(wrong_criterion_config_yaml)
    wrong_criterion_config = Config(wrong_criterion_config_dict)
    wrong_criterion = Component.from_name(
        list(wrong_criterion_config.keys())[0], config=wrong_criterion_config
    )
    datapoint = np.array([0, 1, 2, 3, 4])

    # Verify that a TraitError is raisesd when calling the RangeCriterion component
    with pytest.raises(
        TraitError,
    ) as exc_info:
        wrong_criterion(datapoint)

    # Access the original cause of the TraitError
    original_exception = exc_info.value.__context__
    assert isinstance(original_exception, KeyError)
    assert "'max_value'" in str(original_exception)


@pytest.mark.verifies_usecase("UC-140-1.2.1")
def test_missing_min_value_in_telescope_range_criterion_config():
    """Missing entirely 'min_value: - [type, LST*, 3.]' item"""

    subarray = SubarrayDescription.from_hdf(f"{data_path}/MonteCarloArray.hdf")
    wrong_criterion_config_yaml = """
    TelescopeRangeCriterion:
      max_value:
        - [type, LST*, 10]
        - [type, MST*, 20]
    """
    wrong_criterion_config_dict = yaml.safe_load(wrong_criterion_config_yaml)
    wrong_criterion_config = Config(wrong_criterion_config_dict)
    wrong_criterion = TelescopeComponent.from_name(
        list(wrong_criterion_config.keys())[0],
        subarray,
        config=wrong_criterion_config,
    )
    datapoint = np.array([0, 1, 2, 3, 4])

    # Verify that a KeyError is raised when calling the TelescopeRangeCriterion component
    with pytest.raises(
        KeyError,
        match="TelescopeParameterLookup: no parameter value was set for telescope with tel_id=1. Please set it explicitly, or by telescope type or '*'.",
    ):
        wrong_criterion(datapoint, tel_id=1)


@pytest.mark.verifies_usecase("UC-140-1.2.1")
def test_missing_max_value_in_telescope_range_criterion_config():
    """Missing entirely 'max_value: - [type, LST*, 20]' item"""

    subarray = SubarrayDescription.from_hdf(f"{data_path}/MonteCarloArray.hdf")
    wrong_criterion_config_yaml = """
    TelescopeRangeCriterion:
      min_value:
        - [type, LST*, 3.]
        - [type, MST*, 11]
    """
    wrong_criterion_config_dict = yaml.safe_load(wrong_criterion_config_yaml)
    wrong_criterion_config = Config(wrong_criterion_config_dict)
    wrong_criterion = TelescopeComponent.from_name(
        list(wrong_criterion_config.keys())[0],
        subarray,
        config=wrong_criterion_config,
    )
    datapoint = np.array([12, 13, 14, 15, 16])

    # Verify that a KeyError is raised when calling the TelescopeRangeCriterion component
    with pytest.raises(
        KeyError,
        match="TelescopeParameterLookup: no parameter value was set for telescope with tel_id=1. Please set it explicitly, or by telescope type or '*'.",
    ):
        wrong_criterion(datapoint, tel_id=1)


@pytest.mark.verifies_usecase("UC-140-1.2.1")
def test_missing_threshold_parameter_in_threshold_criterion_config():
    """Missing entirely 'threshold: 10' item."""

    wrong_criterion_config_yaml = """
    ThresholdCriterion:
      above: True
    """
    wrong_criterion_config_dict = yaml.safe_load(wrong_criterion_config_yaml)
    wrong_criterion_config = Config(wrong_criterion_config_dict)
    criterion = Component.from_name(
        list(wrong_criterion_config.keys())[0], config=wrong_criterion_config
    )
    datapoint = np.array([0, 1, 2, 3, 4])

    # Verify that a TraitError is raisesd when calling the ThresholdCriterion component
    with pytest.raises(
        TraitError,
    ) as exc_info:
        criterion(datapoint)

    # Access the original cause of the TraitError
    original_exception = exc_info.value.__context__
    assert isinstance(original_exception, KeyError)
    assert "'threshold'" in str(original_exception)


@pytest.mark.verifies_usecase("UC-140-1.2.1")
def test_missing_above_parameter_in_threshold_criterion_config():
    """Missing entirely 'above: True' item."""

    wrong_criterion_config_yaml = """
    ThresholdCriterion:
      threshold: 10
    """
    wrong_criterion_config_dict = yaml.safe_load(wrong_criterion_config_yaml)
    wrong_criterion_config = Config(wrong_criterion_config_dict)
    wrong_criterion = Component.from_name(
        list(wrong_criterion_config.keys())[0], config=wrong_criterion_config
    )
    datapoint = np.array([0, 1, 2, 3, 4])

    # Verify that a TraitError is raisesd when calling the ThresholdCriterion component
    with pytest.raises(
        TraitError,
        match="The 'above' trait of a ThresholdCriterion instance expected a boolean, not the NoneType None.",
    ) as exc_info:
        wrong_criterion(datapoint)

    # Access the original cause of the TraitError
    original_exception = exc_info.value.__context__
    assert isinstance(original_exception, KeyError)
    assert "'above'" in str(original_exception)


@pytest.mark.verifies_usecase("UC-140-1.2.1")
def test_missing_threshold_parameter_in_telescope_threshold_criterion_config():
    """Missing entirely 'threshold: - [type, LST*, 10]' item"""

    subarray = SubarrayDescription.from_hdf(f"{data_path}/MonteCarloArray.hdf")
    wrong_criterion_config_yaml = """
    TelescopeThresholdCriterion:
      above: True
    """
    wrong_criterion_config_dict = yaml.safe_load(wrong_criterion_config_yaml)
    wrong_criterion_config = Config(wrong_criterion_config_dict)
    wrong_criterion = TelescopeComponent.from_name(
        list(wrong_criterion_config.keys())[0],
        subarray,
        config=wrong_criterion_config,
    )
    datapoint = np.array([0, 1, 2, 3, 4])

    # Verify that a KeyError is raised when calling the TelescopeThresholdCriterion component
    with pytest.raises(
        KeyError,
        match="TelescopeParameterLookup: no parameter value was set for telescope with tel_id=1. Please set it explicitly, or by telescope type or '*'.",
    ):
        wrong_criterion(datapoint, tel_id=1)


@pytest.mark.verifies_usecase("UC-140-1.2.1")
def test_missing_above_parameter_in_telescope_threshold_criterion_config():
    """Missing entirely 'above: True' item"""

    subarray = SubarrayDescription.from_hdf(f"{data_path}/MonteCarloArray.hdf")
    wrong_criterion_config_yaml = """
    TelescopeThresholdCriterion:
      threshold:
        - [type, LST*, 10]
        - [type, MST*, 20]
    """
    wrong_criterion_config_dict = yaml.safe_load(wrong_criterion_config_yaml)
    wrong_criterion_config = Config(wrong_criterion_config_dict)
    wrong_criterion = TelescopeComponent.from_name(
        list(wrong_criterion_config.keys())[0],
        subarray,
        config=wrong_criterion_config,
    )
    datapoint = np.array([0, 1, 2, 3, 4])

    # Verify that a TraitError is raisesd when calling the TelescopeThresholdCriterion component
    with pytest.raises(
        TraitError,
        match="The 'above' trait of a TelescopeThresholdCriterion instance expected a boolean, not the NoneType None.",
    ) as exc_info:
        wrong_criterion(datapoint)

    # Access the original cause of the TraitError
    original_exception = exc_info.value.__context__
    assert isinstance(original_exception, KeyError)
    assert "'above'" in str(original_exception)
